import pytest

from vininfo import Vin, ValidationError


def test_validation():

    with pytest.raises(ValidationError):
         Vin('tooshort')

    with pytest.raises(ValidationError):
         Vin('AAAAAAAAAAAAAAAAO')


def test_basic():

    # number faked
    vin = Vin('JSA12345678901234')
    assert vin.manufacturer == 'Suzuki'
    assert not vin.manufacturer_is_small

    # number faked
    assert Vin('TM912345678901234').manufacturer_is_small


def test_checksum():

    assert Vin('1M8GDM9AXKP042788').verify_checksum()

    # faked
    assert not Vin('1M8GDM9AyKP042788').verify_checksum()


def test_unsupported_brand():

    vin = Vin('200BL8EV9AX604020')
    assert vin.manufacturer == 'UnsupportedBrand'
    assert vin.country == 'Canada'


def test_merge_wmi():
    from vininfo.utils import merge_wmi

    missing, lines = merge_wmi({'1DTEST': 'Some', '1GTEST': 'Other'})
    assert missing == {'1DTEST', '1GTEST'}
    assert "    '1D': 'Dodge',\n    '1DTEST': 'Some'," in lines
    assert "    '1GT': 'GMC Truck',\n    '1GTEST': 'Other'," in lines



# -*- encoding: utf-8 -*-
from __future__ import unicode_literals

from ..common import Annotatable

if False:  # pragma: nocover
    from ..toolbox import Vin


class DetailWrapper(object):

    __slots__ = ['code', 'name']

    def __init__(self, details, detail):
        """
        :param VinDetails details:
        :param Detail detail: 
        """
        vin = details._vin
        attr_name, attr_idx = detail.source
        code_source = getattr(vin, attr_name)

        code = code_source[attr_idx]

        defs = detail.defs

        if callable(defs):
            defs = defs(details)

        self.code = code
        self.name = defs.get(code)

    def __str__(self):
        return self.name or self.code


class Detail(object):
    """Vin detail descriptor."""

    __slots__ = ['source', 'defs']

    def __init__(self, code_source=None, definitions=None):
        self.source = code_source
        self.defs = definitions or {}

    def __get__(self, instance, owner):
        """
        :param VinDetails instance:
        :param owner:
        :rtype: DetailWrapper
        """

        return DetailWrapper(instance, self)


class VinDetails(Annotatable):
    """Offers advanced (manufacturer specific) VIN data extraction ficilities."""

    annotate_titles = {
        'body': 'Body',
        'engine': 'Engine',
        'model': 'Model',
        'plant': 'Plant',
        'transmission': 'Transmission',
        'serial': 'Serial',
    }

    def __init__(self, vin):
        """
        :param Vin vin:
        """
        self._vin = vin
        # self._cached = {}

    body = Detail()
    engine = Detail()
    model = Detail()
    plant = Detail()
    transmission = Detail()
    serial = Detail()



# -*- encoding: utf-8 -*-
from __future__ import unicode_literals

from ._base import VinDetails, Detail
from ..dicts.bodies import *


class AvtoVazDetails(VinDetails):
    """AvtoVAZ VIN details extractor."""

    model = Detail(('vds', 1), {
        'A': 'XRAY',
        'F': 'Vesta',
    })

    body = Detail(('vds', 2), {
        'B': BODY_HATCH_5,
        'K': BODY_SW_5,
        'L': BODY_SEDAN_4,
    })

    engine = Detail(('vds', 3), {
        '1': '21129',
        '2': '11189',
        '3': '21179',
        '4': 'H4M',
        'A': '21129 CNG',
    })

    plant = Detail(('vis', 1), {
        '0': 'Tolyatti',
        'Y': 'Izhevsk',
    })

    transmission = Detail(('vds', 4), {
        '1': 'Manual, 5-Gear (VAZ 21807)',
        '2': 'Semi-automatic, 5-Gear (VAZ 21827)',
        '3': 'Manual, 5-Gear (Renault JH3 514)',
    })

    serial = Detail(('vis', slice(2, None)))



# -*- encoding: utf-8 -*-
from __future__ import unicode_literals


def __unpack_countries_map(counties):
    unpacked = {}

    seq = 'ABCDEFGHJKLMNPRSTUVWXYZ1234567890'

    for code, title in counties.items():
        first, _, span  = code.partition('-')

        if span:
            ch_from, ch_till = span

        else:
            ch_from = 'A'
            ch_till = '0'

        for ch in seq[seq.index(ch_from):seq.index(ch_till) + 1]:
            unpacked[first + ch] = title

    return unpacked


COUNTRIES = __unpack_countries_map({
    'A-AH': 'South Africa',
    'A-JN': "Cote d'Ivoire",
    'B-AE': 'Angola',
    'B-FK': 'Kenya',
    'B-LR': 'Tanzania',
    'C-AE': 'Benin',
    'C-FK': 'Madagascar',
    'C-LR': 'Tunisia',
    'D-AE': 'Egypt',
    'D-FK': 'Morocco',
    'D-LR': 'Zambia',
    'E-AE': 'Ethiopia',
    'E-FK': 'Mozambique',
    'F-AE': 'Ghana',
    'F-FK': 'Nigeria',
    'J-': 'Japan',
    'K-AE': 'Sri Lanka',
    'K-FK': 'Israel',
    'K-LR': 'Korea (South)',
    'K-S0': 'Kazakhstan',
    'L-': 'China (Mainland)',
    'M-AE': 'India',
    'M-FK': 'Indonesia',
    'M-LR': 'Thailand',
    'M-S0': 'Myanmar',
    'N-AE': 'Iran',
    'N-FK': 'Pakistan',
    'N-LR': 'Turkey',
    'P-AE': 'Philippines',
    'P-FK': 'Singapore',
    'P-LR': 'Malaysia',
    'R-AE': 'United Arab Emirates',
    'R-FK': 'Taiwan, China',
    'R-LR': 'Vietnam',
    'R-S0': 'Saudi Arabia',
    'S-AM': 'United Kingdom',
    'S-NT': 'Germany/East Germany',
    'S-UZ': 'Poland',
    'S-14': 'Latvia',
    'T-AH': 'Switzerland',
    'T-JP': 'Czech Republic',
    'T-RV': 'Hungary',
    'T-W1': 'Portugal',
    'U-HM': 'Denmark',
    'U-NT': 'Ireland',
    'U-UZ': 'Romania',
    'U-57': 'Slovakia',
    'V-AE': 'Austria',
    'V-FR': 'France',
    'V-SW': 'Spain',
    'V-X2': 'Serbia',
    'V-35': 'Croatia',
    'V-60': 'Estonia',
    'W-': 'Germany/West Germany',
    'X-AE': 'Bulgaria',
    'X-FK': 'Greece',
    'X-LR': 'Netherlands',
    'X-SW': 'USSR/CIS',
    'X-X2': 'Luxembourg',
    'X-30': 'Russia',
    'Y-AE': 'Belgium',
    'Y-FK': 'Finland',
    'Y-LR': 'Malta',
    'Y-SW': 'Sweden',
    'Y-X2': 'Norway',
    'Y-35': 'Belarus',
    'Y-60': 'Ukraine',
    'Z-AR': 'Italy',
    'Z-X2': 'Slovenia',
    'Z-35': 'Lithuania',
    '1-': 'United States',
    '2-': 'Canada',
    '3-AW': 'Mexico',
    '3-X7': 'Costa Rica',
    '3-89': 'Cayman Islands',
    '4-': 'United States',
    '5-': 'United States',
    '6-': 'Australia',
    '7-': 'New Zealand',
    '8-AE': 'Argentina',
    '8-FK': 'Chile',
    '8-LR': 'Ecuador',
    '8-SW': 'Peru',
    '8-X2': 'Venezuela',
    '9-AE': 'Brazil',
    '9-FK': 'Colombia',
    '9-LR': 'Paraguay',
    '9-SW': 'Uruguay',
    '9-X2': 'Trinidad & Tobago',
    '9-39': 'Brazil',
})



# -*- encoding: utf-8 -*-
from __future__ import unicode_literals

from collections import OrderedDict


class Annotatable(object):

    annotate_titles = {}

    def annotate(self):

        annotations = {}
        no_attr = set()

        for attr_name, label in self.annotate_titles.items():
            value = getattr(self, attr_name, no_attr)

            if value is no_attr:
                continue

            if isinstance(value, list):
                value = ', '.join('%s' % val for val in value)

            annotations[label] = '%s' % value

        return OrderedDict((title, value) for title, value in sorted(annotations.items(), key=lambda item: item[0]))


class Brand(object):

    __slots__ = ['manufacturer']

    extractor = None

    def __init__(self, manufacturer=None):
        self.manufacturer = manufacturer or self.title

    @property
    def title(self):
        return self.__class__.__name__

    def __str__(self):
        return '%s (%s)' % (self.title, self.manufacturer)


class UnsupportedBrand(Brand):
    """Unsupported brand."""



from vininfo import Vin


def test_nissan():

    vin = Vin('5N1NJ01CXST000001')

    assert '%s' % vin
    assert vin.wmi == '5N1'
    assert vin.manufacturer == 'Nissan'
    assert vin.vds == 'NJ01CX'
    assert vin.vis == 'ST000001'
    assert vin.years == [1995]
    assert vin.region_code == '5'
    assert vin.region == 'North America'
    assert vin.country_code == '5N'
    assert vin.country == 'United States'
    assert '%s' % vin.brand == 'Nissan (Nissan)'

    details = vin.details
    assert details.model.code == 'J'
    assert details.model.name == 'Maxima'
    assert details.body.code == '1'
    assert details.body.name == ['Sedan, 4-Door', 'Standard Body Truck']
    assert details.engine.code == 'N'
    assert details.engine.name == 'VH45DE'
    assert details.plant.code == 'T'
    assert details.plant.name == ['Tochigi', 'Oppama']
    assert details.serial.code == '000001'

    assert vin.verify_checksum()



# -*- encoding: utf-8 -*-
from __future__ import unicode_literals


BODY_CABRI_2 = 'Cabriolet, 2-Door'
BODY_COUPE_2 = 'Coupe, 2-Door'
BODY_CROSS_3 = 'Crossover, 3-Door'
BODY_PICKUP_2 = 'Pickup, 2-Door'

BODY_HATCH_3 = 'Hatchback, 3-Door'
BODY_HATCH_5 = 'Hatchback, 5-Door'

BODY_MINIVAN_3 = 'Minivan, 3-Door'
BODY_MINIVAN_5 = 'Minivan, 5-Door'

BODY_SEDAN_4 = 'Sedan, 4-Door'
BODY_SEDAN_2 = 'Sedan, 2-Door'

BODY_SW_3 = 'Station Wagon, 3-Door'
BODY_SW_5 = 'Station Wagon, 5-Door'
BODY_SW_8 = 'Station Wagon, 8-Door'

BODY_MINIBUS = 'Minibus'
BODY_VAN = 'Van'



from .toolbox import Vin
from .exceptions import ValidationError, VininfoException


VERSION = (1, 3, 0)
"""Application version number tuple."""

VERSION_STR = '.'.join(map(str, VERSION))
"""Application version number string."""


# -*- encoding: utf-8 -*-
from __future__ import unicode_literals

from datetime import datetime
from itertools import cycle

from .common import Annotatable, Brand, UnsupportedBrand
from .dicts import COUNTRIES, WMI, REGIONS
from .exceptions import ValidationError
from .utils import string_types


if False:  # pragma: nocover
    from .details._base import VinDetails


class Vin(Annotatable):
    """Offers basic VIN data extraction facilities."""

    annotate_titles = {
        'manufacturer': 'Manufacturer',
        'region': 'Region',
        'country': 'Country',
        'years': 'Years',
    }

    def __init__(self, num):
        self.num = self.validate(num)

        details_extractor = self.brand.extractor

        if details_extractor:
            details_extractor = details_extractor(self)

        self.details = details_extractor  # type: VinDetails

    def __str__(self):
        return self.num

    @classmethod
    def validate(self, num):
        """Performs basic VIN validation and sanation.

        :param str|unicode num:
        :rtype: str|unicode
        """
        num = num.strip().upper()

        num_len = len(num)
        if num_len != 17:
            raise ValidationError('VIN number requires 17 chars (%s given)' % num_len)

        illegal = {'I', 'O', 'Q'}

        for ch in num:
            if ch in illegal:
                raise ValidationError('VIN number should not contain: %s' % ', '.join(illegal))

        return num

    def verify_checksum(self):
        """Performs checksum verification.

        .. warning:: Not every manufacturer uses VIN checksum rules.

        :rtype: bool

        """
        if self.vis[0] in {'U', 'Z', '0'}:
            return False

        trans = {
            'A': 1, 'B': 2, 'C': 3, 'D': 4, 'E': 5, 'F': 6, 'G': 7, 'H': 8,
            'J': 1, 'K': 2, 'L': 3, 'M': 4, 'N': 5,         'P': 7,         'R': 9,
                    'S': 2, 'T': 3, 'U': 4, 'V': 5, 'W': 6, 'X': 7, 'Y': 8, 'Z': 9,
        }
        weights = (8, 7, 6, 5, 4, 3, 2, 10, 0, 9, 8, 7, 6, 5, 4, 3, 2)

        checksum = 0

        for pos, char in enumerate(self.num):
            value = int(trans.get(char, char))
            checksum += (value * weights[pos])

        checksum = int(checksum) % 11

        check_digit = 'X' if checksum == 10 else checksum

        return str(check_digit) == self.vds[5]

    @property
    def wmi(self):
        """WMI (World Manufacturers Identification)"""
        return self.num[:3]

    @property
    def brand(self):
        """Brand object.

        :rtype: Brand
        """
        wmi = self.wmi

        brand = WMI.get(wmi)

        if not brand:
            brand = WMI.get(wmi[:2])

        if isinstance(brand, string_types):
            brand = Brand(brand)

        if brand is None:
            brand = UnsupportedBrand()

        return brand

    @property
    def manufacturer(self):
        """Manufacturer title.

        :rtype: str|unicode
        """
        return self.brand.manufacturer

    @property
    def manufacturer_is_small(self):
        """A manufacturer who builds fewer than 1000 vehicles per year.

        :rtype: bool
        """
        return str(self.wmi[2]) == '9'

    @property
    def vds(self):
        """VDS (Vehicle Description Section)"""
        return self.num[3:9]

    @property
    def vis(self):
        """VIS (Vehicle Identifier Section)"""
        return self.num[9:17]

    @property
    def region_code(self):
        return self.wmi[0]

    @property
    def region(self):
        code = self.region_code

        title = None

        for chars, title_ in REGIONS.items():
            if code in chars:
                title = title_
                break

        return title

    @property
    def country_code(self):
        return self.wmi[0:2]

    @property
    def country(self):
        return COUNTRIES.get(self.country_code)

    @property
    def years(self):
        letters = 'ABCDEFGHJKLMNPRSTVWXY123456789'
        year_letter = self.vis[0]

        year = 1979
        year_current = datetime.now().year

        result = []

        for letter in cycle(letters):
            year += 1

            if letter == year_letter:
                result.append(year)

            if year == year_current:
                break

        result.sort(reverse=True)

        return result



# -*- encoding: utf-8 -*-
from __future__ import unicode_literals


class VininfoException(Exception):
    """Base exception."""


class ValidationError(VininfoException):
    """Data validation exception."""



# -*- encoding: utf-8 -*-
from __future__ import unicode_literals


REGIONS = {
    'ABCDEFGH': 'Africa',
    'JKLMNPR': 'Asia',
    'STUVWXYZ': 'Europe',
    '12345': 'North America',
    '67': 'Oceania',
    '89': 'South America',
}
